import time
import unittest

from modela.ModelaException import *
from modela.training.Study import Study
from modela.server import Modela
from modela.training.models import StudySpec, ModelSearch, Ensemble, Training, DataSplit
from modela import *


class Test_Modela_study(unittest.TestCase):
    """Tests for `modela.training.Study`"""

    @classmethod
    def setUpClass(cls):
        cls.modela = Modela(port_forward=True)

    @classmethod
    def tearDownClass(cls):
        cls.modela.close()

    def test_0_create(self):
        study = self.modela.Study(namespace="iris-product", name="test")
        study.spec.Location = DataLocation(BucketName="test")
        assert type(study) == Study
        # study.submit(replace=True)
        Study(client=self.modela.Studies,
              name="test",
              namespace="iris-product",
              dataset="iris",
              task_type=TaskType.MultiClassification,
              objective=Metric.Accuracy,
              bucket="default-minio-bucket",
              search=ModelSearch(MaxTime=200, MaxModels=2, Trainers=2,
                                 SearchSpace=AlgorithmSearchSpace(
                                     Allowlist=[ClassicEstimator.LogisticRegression]))).submit(replace=True)

    def test_1_list(self):
        assert len(self.modela.Studies.list("iris-product")) >= 1

    def test_2_update(self):
        study = self.modela.Study(namespace="iris-product", name="test")
        study.set_label("test", "e")
        study.update()
        newds = self.modela.Study(namespace="iris-product", name="test")
        assert newds.has_label("test")

    def test_3_get(self):
        study = self.modela.Study(namespace="iris-product", name="test")
        spec = study.spec
        assert type(spec) == StudySpec

    def test_4_delete(self):
        study = self.modela.Study(namespace="iris-product", name="test")
        study.delete()
        self.assertRaises(ResourceNotFoundException, self.modela.Studies.get, "iris-product", "test")

    def test_5_get_models(self):
        study = self.modela.Studies.list("iris-product")[0]
        assert len(study.models) > 0


""" These tests should be run by an end-user. """


def test_viz():
    modela = Modela("localhost", 3000)
    study = modela.Study(
        name="test-study-2",
        namespace="iris-product",
        dataset="iris",
        task_type=TaskType.MultiClassification,
        objective=Metric.Accuracy,
        bucket="default-minio-bucket",
        search=ModelSearch(MaxTime=200, MaxModels=8, Trainers=4))

    study.submit_and_visualize(replace=True)
    modela.close()



from dataclasses import dataclass, field
from typing import List

from github.com.metaprov.modelaapi.pkg.apis.inference.v1alpha1.generated_pb2 import \
    PredictorCondition as MDPredictorCondition
from github.com.metaprov.modelaapi.pkg.apis.catalog.v1alpha1.generated_pb2 import \
    ModelDeploymentSpec as MDModelDeploymentSpec
from github.com.metaprov.modelaapi.pkg.apis.inference.v1alpha1.generated_pb2 import ModelRecord as MDModelRecord
from github.com.metaprov.modelaapi.pkg.apis.inference.v1alpha1.generated_pb2 import MonitorStatus as MDMonitorStatus
from github.com.metaprov.modelaapi.pkg.apis.catalog.v1alpha1.generated_pb2 import \
    ModelDeploymentStatus as MDModelDeploymentStatus
from github.com.metaprov.modelaapi.pkg.apis.catalog.v1alpha1.generated_pb2 import RunSchedule as MDRunSchedule
from github.com.metaprov.modelaapi.services.grpcinferenceservice.v1.grpcinferenceservice_pb2 import \
    PredictRequest as MDPredictRequest, ProbabilityValue as MDProbabilityValue, ShapValue as MDShapValue

from modela.common import Configuration, ConditionStatus, Time, ObjectReference, StatusError, TriggerScheduleEventType
from modela.inference.common import PredictorConditionType, CanaryMetric, AccessType, PredictorType, \
    ModelDeploymentPhase
from modela.infra.models import Workload
from modela.training.common import TaskType
from modela.training.models import ModelValidation, ModelValidationResult


@dataclass
class RunSchedule(Configuration):
    Enabled: bool = False
    StartTime: Time = None
    EndTime: Time = None
    Cron: str = ""
    Type: TriggerScheduleEventType = None

    def to_message(self) -> MDRunSchedule:
        return self.set_parent(MDRunSchedule()).parent


@dataclass
class ModelDeploymentStatus(Configuration):
    ImageName: str = ""
    DeploymentRef: ObjectReference = None
    ServiceRef: ObjectReference = None
    HpaName: str = ""
    Current95: float = 0
    Current99: float = 0
    LastPredictionTime: Time = None
    DailyPredictionAvg: int = 0
    LastFailure: str = ""
    Phase: ModelDeploymentPhase = None
    DeployedAt: Time = None
    ReleasedAt: Time = None
    TrainingDatasetName: str = ""
    ApprovedBy: str = ""
    ApprovedAt: Time = None

    def to_message(self) -> MDModelDeploymentStatus:
        return self.set_parent(MDModelDeploymentStatus()).parent


@dataclass
class ModelRecord(Configuration):
    ModelName: str = ""
    ModelVersion: str = ""
    LiveAt: Time = None
    RetiredAt: Time = None
    AvgDailyPrediction: int = 0
    AvgLatency: float = 0

    def to_message(self) -> MDModelRecord:
        return self.set_parent(MDModelRecord()).parent


@dataclass
class ModelDeploymentSpec(Configuration):
    ModelName: str = ""
    ModelVersion: str = ""
    MaxTraffic: int = 0
    Traffic: int = 0
    Canary: bool = False
    Shadow: bool = False
    Released: bool = False
    Deployed: bool = False
    MountTar: bool = False
    TrafficSelector: str = ""
    CanaryMetrics: List[CanaryMetric] = field(default_factory=lambda: [])

    def to_message(self) -> MDModelDeploymentSpec:
        return self.set_parent(MDModelDeploymentSpec()).parent


@dataclass
class PredictorCondition(Configuration):
    Type: PredictorConditionType = None
    Status: ConditionStatus = None
    LastTransitionTime: Time = None
    Reason: str = ""
    Message: str = ""

    def to_message(self) -> MDPredictorCondition:
        return self.set_parent(MDPredictorCondition()).parent


@dataclass
class ProgressiveSpec(Configuration):
    Warmup: int = 0
    TrafficIncrement: int = 0
    CanaryMetrics: List[CanaryMetric] = field(default_factory=lambda: [])


@dataclass
class PredictionCacheSpec(Configuration):
    Active: bool = False
    ServiceName: str = ""


@dataclass
class AutoScaling(Configuration):
    Enabled: bool = False
    MinReplicas: int = 1
    MaxReplicas: int = 1
    CpuAvgUtilization: int = 80
    MemAvgUtilization: int = 80


@dataclass
class PredictorHealth(Configuration):
    Service: bool = False
    DataDrift: bool = False
    ConceptDrift: bool = False
    TotalPredictions: int = 0
    Avg: int = 0
    TotalP95Requests: int = 0
    MedianResponseTime: int = 0
    LastDailyPredictions: List[int] = field(default_factory=lambda: [])


@dataclass
class PredictorStatus(Configuration):
    ModelStatus: List[ModelDeploymentStatus] = field(default_factory=lambda: [])
    MonitorLastAttemptAt: Time = None
    MonitorLastScore: float = 0
    MonitorLastLatency: float = 0
    Health: PredictorHealth = None
    ObservedGeneration: int = 0
    History: List[ModelRecord] = field(default_factory=lambda: [])
    MonitorStatus: MonitorStatus = None
    LastUpdated: Time = None
    TargetColumn: str = ""
    PositiveLabel: str = ""
    NegativeLabel: str = ""
    EndPoint: str = ""
    ProxyDeploymentRef: ObjectReference = None
    ProxyServiceRef: ObjectReference = None
    FailureReason: StatusError = None
    FailureMessage: str = ""
    Conditions: List[PredictorCondition] = field(default_factory=lambda: [])


@dataclass
class MonitorSpec(Configuration):
    Enabled: bool = False
    SamplePercent: int = 0
    Schedule: RunSchedule = None
    NotifierName: str = ""
    Validations: List[ModelValidation] = field(default_factory=lambda: [])
    LogRequests: bool = False
    LogResponses: bool = False


@dataclass
class MonitorStatus(Configuration):
    LastPrediction: Time = None
    ValidationResults: List[ModelValidationResult] = field(default_factory=lambda: [])

    def to_message(self) -> MDMonitorStatus:
        return self.set_parent(MDMonitorStatus()).parent


@dataclass
class PredictorAuthSpec(Configuration):
    Enabled: bool = False


PredictorAccessType = AccessType
PredictorAutoScaling = AutoScaling


@dataclass
class PredictorSpec(Configuration):
    VersionName: str = ""
    Description: str = ""
    ProductRef: ObjectReference = None
    ServingsiteRef: ObjectReference = ObjectReference("default-tenant", "default-lab")
    # Template: PodTemplate = None
    Models: List[ModelDeploymentSpec] = field(default_factory=lambda: [])
    Progressive: ProgressiveSpec = None
    ArtifactsFolder: str = ""
    Port: int = 8080
    NodePort: int = 30000
    Path: str = ""
    AccessType: PredictorAccessType = None
    Replicas: int = 1
    AutoScaling: PredictorAutoScaling = None
    Owner: str = "no-one"
    Resources: Workload = Workload("general-nano")
    Cache: PredictionCacheSpec = None
    # Store: OnlineFeaturestoreSpec = None
    ForwardCurtain: str = ""
    BackwardCurtain: str = ""
    Type: PredictorType = PredictorType.Online
    Task: TaskType = None
    PredictionThreshold: float = 0
    Monitor: MonitorSpec = None
    Auth: PredictorAuthSpec = None

@dataclass
class ProbabilityValue(Configuration):
    Label: str = ""
    Probability: float = 0

    def to_message(self) -> MDProbabilityValue:
        return self.set_parent(MDProbabilityValue()).parent

@dataclass
class ShapValue(Configuration):
    Feature: str = ""
    Value: float = 0

    def to_message(self) -> MDShapValue:
        return self.set_parent(MDShapValue()).parent


@dataclass
class PredictionResult(Configuration):
    Success: bool = False
    Score: float = 0
    Label: str = ""
    Probabilities: List[ProbabilityValue] = field(default_factory=lambda: [])
    ShapValues: List[ShapValue] = field(default_factory=lambda: [])
    MissingColumns: List[str] = field(default_factory=lambda: [])
    OutOfBound: List[str] = field(default_factory=lambda: [])
    BaseShapValue: float = 0






import unittest

from modela.ModelaException import *
from modela.inference.Predictor import Predictor
from modela.inference.models import PredictorSpec
from modela.server import Modela
from modela import *


class Test_Modela_predictor(unittest.TestCase):
    """Tests for `modela.inference.Predictor`"""

    @classmethod
    def setUpClass(cls):
        cls.modela = Modela(port_forward=True)

    @classmethod
    def tearDownClass(cls):
        cls.modela.close()

    def test_0_create(self):
        predictor = self.modela.Predictor(namespace="iris-product", name="test")
        predictor.submit(replace=True)

    def test_1_list(self):
        assert len(self.modela.Predictors.list("iris-product")) >= 1

    def test_2_update(self):
        predictor = self.modela.Predictor(namespace="iris-product", name="test")
        predictor.set_label("test", "e")
        predictor.update()
        newds = self.modela.Predictor(namespace="iris-product", name="test")
        assert newds.has_label("test")

    def test_3_get(self):
        predictor = self.modela.Predictor(namespace="iris-product", name="test")
        spec = predictor.spec
        assert type(spec) == PredictorSpec

    def test_4_delete(self):
        predictor = self.modela.Predictor(namespace="iris-product", name="test")
        predictor.delete()
        self.assertRaises(ResourceNotFoundException, self.modela.Studies.get, "iris-product", "test")

    def test_5_predict(self):
        predictor = self.modela.Predictor(namespace="iris-product", name="testpred", model="model-20220303-194814")
        predictor.submit(replace=True)
        predictor.wait_until_ready()
        predictor.predict(predictor.model.test_prediction)
        predictor.delete()




import grpc
from github.com.metaprov.modelaapi.pkg.apis.training.v1alpha1.generated_pb2 import ModelAutobuilder as MDModelAutobuilder
from github.com.metaprov.modelaapi.services.modelautobuilder.v1.modelautobuilder_pb2_grpc import ModelAutobuilderServiceStub
from github.com.metaprov.modelaapi.services.modelautobuilder.v1.modelautobuilder_pb2 import CreateModelAutobuilderRequest, \
    UpdateModelAutobuilderRequest, \
    DeleteModelAutobuilderRequest, GetModelAutobuilderRequest, ListModelAutobuildersRequest

from modela.Resource import Resource
from modela.ModelaException import ModelaException
from typing import List, Union


class ModelAutobuilder(Resource):
    def __init__(self, item: MDModelAutobuilder = MDModelAutobuilder(), client=None, namespace="", name="", version=Resource.DefaultVersion):
        super().__init__(item, client, namespace=namespace, name=name, version=version)


class ModelAutobuilderClient:
    def __init__(self, stub, modela):
        self.modela = modela
        self.__stub: ModelAutobuilderServiceStub = stub

    def create(self, modelautobuilder: ModelAutobuilder) -> bool:
        request = CreateModelAutobuilderRequest()
        request.modelautobuilder.CopyFrom(modelautobuilder.raw_message)
        try:
            response = self.__stub.CreateModelAutobuilder(request)
            return True
        except grpc.RpcError as err:
            error = err

        ModelaException.process_error(error)
        return False

    def update(self, modelautobuilder: ModelAutobuilder) -> bool:
        request = UpdateModelAutobuilderRequest()
        request.modelautobuilder.CopyFrom(modelautobuilder.raw_message)
        try:
            self.__stub.UpdateModelAutobuilder(request)
            return True
        except grpc.RpcError as err:
            error = err

        ModelaException.process_error(error)
        return False

    def get(self, namespace: str, name: str) -> Union[ModelAutobuilder, bool]:
        request = GetModelAutobuilderRequest()
        request.namespace = namespace
        request.name = name
        try:
            response = self.__stub.GetModelAutobuilder(request)
            return ModelAutobuilder(response.modelautobuilder, self)
        except grpc.RpcError as err:
            error = err

        ModelaException.process_error(error)
        return False

    def delete(self, namespace: str, name: str) -> bool:
        request = DeleteModelAutobuilderRequest()
        request.namespace = namespace
        request.name = name
        try:
            response = self.__stub.DeleteModelAutobuilder(request)
            return True
        except grpc.RpcError as err:
            error = err

        ModelaException.process_error(error)
        return False

    def list(self, namespace: str, labels: dict = None) -> Union[List[ModelAutobuilder], bool]:
        request = ListModelAutobuildersRequest()
        request.namespace = namespace
        if labels is not None:
            request.labels.update(labels)

        try:
            response = self.__stub.ListModelAutobuilders(request)
            return [ModelAutobuilder(item, self) for item in response.modelautobuilders.items]
        except grpc.RpcError as err:
            error = err

        ModelaException.process_error(error)
        return False





from dataclasses import dataclass, field
from typing import List

from github.com.metaprov.modelaapi.pkg.apis.catalog.v1alpha1.generated_pb2 import Measurement as MDMeasurement
from github.com.metaprov.modelaapi.pkg.apis.training.v1alpha1.generated_pb2 import ModelCondition as MDModelCondition
from github.com.metaprov.modelaapi.pkg.apis.training.v1alpha1.generated_pb2 import \
    HyperParameterValue as MDHyperParameterValue
from github.com.metaprov.modelaapi.pkg.apis.training.v1alpha1.generated_pb2 import \
    FeatureEngineeringPipeline as MDFeatureEngineeringPipeline
from github.com.metaprov.modelaapi.pkg.apis.training.v1alpha1.generated_pb2 import \
    FeatureImportance as MDFeatureImportance
from github.com.metaprov.modelaapi.pkg.apis.training.v1alpha1.generated_pb2 import StudyCondition as MDStudyCondition
from github.com.metaprov.modelaapi.pkg.apis.training.v1alpha1.generated_pb2 import \
    TextPipelineSpec as MDTextPipelineSpec
from github.com.metaprov.modelaapi.pkg.apis.training.v1alpha1.generated_pb2 import \
    ResourceConsumption as MDResourceConsumption
from github.com.metaprov.modelaapi.pkg.apis.training.v1alpha1.generated_pb2 import DataHashes as MDDataHashes
from github.com.metaprov.modelaapi.pkg.apis.training.v1alpha1.generated_pb2 import \
    GeneratedColumnSpec as MDGeneratedColumnSpec
from github.com.metaprov.modelaapi.pkg.apis.training.v1alpha1.generated_pb2 import FeaturePair as MDFeaturePair
from github.com.metaprov.modelaapi.pkg.apis.training.v1alpha1.generated_pb2 import SegmentSpec as MDSegmentSpec
from github.com.metaprov.modelaapi.pkg.apis.training.v1alpha1.generated_pb2 import ReportCondition as MDReportCondition
from github.com.metaprov.modelaapi.pkg.apis.training.v1alpha1.generated_pb2 import \
    ModelAutobuilderCondition as MDModelAutobuilderCondition
from github.com.metaprov.modelaapi.pkg.apis.training.v1alpha1.generated_pb2 import ModelValidation as MDModelValidation
from github.com.metaprov.modelaapi.pkg.apis.training.v1alpha1.generated_pb2 import ModelValidationResult as MDModelValidationResult
from modela.Configuration import Configuration, ImmutableConfiguration
from modela.common import PriorityLevel, Time, StatusError, ConditionStatus, ObjectReference, Freq, Plot
from modela.data.common import DataType
from modela.data.models import DataLocation, GovernanceSpec, CompilerSettings, Correlation, DataSourceSpec
from modela.inference.common import AccessType
from modela.infra.models import Workload, OutputLogs, NotificationSetting
from modela.training.common import *


@dataclass
class ModelValidation(Configuration):
    Type: ModelValidationType = None
    PrevModel: str = ""
    DatasetName: str = ""
    DriftFreq: Freq = None
    DriftInterval: int = 0
    Column: str = ""
    Metric: Metric = None
    Min: float = 0
    Max: float = 0
    MinPercent: float = 0
    MaxPercent: float = 0
    Agg: Aggregate = None

    def to_message(self) -> MDModelValidation:
        return self.set_parent(MDModelValidation()).parent


@dataclass
class ModelValidationResult(Configuration):
    Type: str = ""
    DatasetName: str = ""
    ModelName: str = ""
    Column: str = ""
    Error: str = ""
    Metric: Metric = None
    ActualValue: float = 0
    Passed: bool = False
    At: Time = None
    DurationInSec: int = 0


    def to_message(self) -> MDModelValidationResult:
        return self.set_parent(MDModelValidationResult()).parent

@dataclass
class Measurement(ImmutableConfiguration):
    Metric: Metric = Metric.Null
    Value: float = 0

    def to_message(self) -> MDMeasurement:
        return self.set_parent(MDMeasurement()).parent


@dataclass
class Segment(Configuration):
    ColumnName: str = ""
    Op: Operation = Operation.EQ
    Value: str = ""

    def to_message(self) -> MDSegmentSpec:
        return self.set_parent(MDSegmentSpec()).parent


@dataclass
class HyperParameterValue(ImmutableConfiguration):
    Name: str = ""
    Value: str = ""

    def to_message(self) -> MDHyperParameterValue:
        return self.set_parent(MDHyperParameterValue()).parent


@dataclass
class ClassicalEstimatorSpec(ImmutableConfiguration):
    AlgorithmName: str = ""
    Parameters: List[HyperParameterValue] = field(default_factory=lambda: [])


@dataclass
class ChatbotEstimatorSpec(ImmutableConfiguration):  # Not Implemented
    Base: str = ""


@dataclass
class NLPEstimatorSpec(ImmutableConfiguration):  # Not Implemented
    Base: str = ""


@dataclass
class FeatureImportance(ImmutableConfiguration):
    Feature: str = ""
    Importance: float = 0

    def to_message(self) -> MDFeatureImportance:
        return self.set_parent(MDFeatureImportance()).parent


@dataclass
class SuccessiveHalving(ImmutableConfiguration):
    Budget: int = 0
    Bracket: int = 0
    Rung: int = 0
    ConfID: int = 0
    Modality: ModalityType = None


@dataclass
class DataSplit(Configuration):
    Method: DataSplitMethod = DataSplitMethod.Auto
    Train: int = 80
    Validation: int = 0
    Test: int = 20
    SplitColumn: str = ""
    Segments: List[Segment] = field(default_factory=lambda: [])


@dataclass
class Training(Configuration):
    Priority: PriorityLevel = PriorityLevel.Medium
    Cvtype: CvType = CvType.CVTypeKFold
    CV: bool = True
    Folds: int = 5
    Split: DataSplit = DataSplit()
    EvalMetrics: List[Metric] = field(default_factory=lambda: [])
    EarlyStop: bool = False
    CheckpointInterval: int = 10
    Sh: SuccessiveHalving = None
    Seed: float = 42
    Resources: Workload = Workload("general-large")
    Gpu: bool = False
    Distributed: bool = False
    NodeCount: int = 1
    SamplePct: int = 100


@dataclass
class ServingSpec(ImmutableConfiguration):
    Resources: Workload = None


@dataclass
class TextPipelineSpec(ImmutableConfiguration):
    Encoder: TextEncoding = TextEncoding.Auto
    Tokenizer: str = ""
    Stopwords: bool = True
    Pos: bool = True
    Lemma: bool = True
    Stem: bool = True
    Embedding: str = ""
    Svd: bool = True
    MaxSvdComponents: int = 0

    def to_message(self) -> MDTextPipelineSpec:
        return self.set_parent(MDTextPipelineSpec()).parent


@dataclass
class ImagePipelineSpec(ImmutableConfiguration):  # Not Implemented
    Featurizer: str = None


@dataclass
class VideoPipelineSpec(ImmutableConfiguration):  # Not Implemented
    Featurizer: str = None


@dataclass
class AudioPipelineSpec(ImmutableConfiguration):  # Not Implemented
    Featurizer: str = None


@dataclass
class ResourceConsumption(ImmutableConfiguration):
    Cpu: float = 0
    Mem: float = 0
    Gpu: float = 0

    def to_message(self) -> MDResourceConsumption:
        return self.set_parent(MDResourceConsumption()).parent


@dataclass
class DataHashes(ImmutableConfiguration):
    TrainHash: str = ""
    TestingHash: str = ""
    ValidationHash: str = ""

    def to_message(self) -> MDDataHashes:
        return self.set_parent(MDDataHashes()).parent


@dataclass
class GeneratedColumnSpec(ImmutableConfiguration):
    Name: str = ""
    Datatype: DataType = DataType.Text
    First: str = ""
    Second: str = ""
    Original: str = ""

    def to_message(self) -> MDGeneratedColumnSpec:
        return self.set_parent(MDGeneratedColumnSpec()).parent


@dataclass
class FeatureSelection(Configuration):
    Enabled: bool = False
    SamplePct: int = 100
    Embedding: bool = False
    Filter: bool = False
    Wrapper: bool = False
    Pipeline: List[FeatureSelectionType] = field(default_factory=lambda: [])
    VarianceThresholdPct: int = 5
    CorrThreshold: int = 95
    TopN: int = 0
    CumulativeImportancePercent: int = 95
    Reserved: List[str] = field(default_factory=lambda: [])


@dataclass
class FeaturePair(ImmutableConfiguration):
    X: str = ""
    Y: str = ""

    def to_message(self) -> MDFeaturePair:
        return self.set_parent(MDFeaturePair()).parent


@dataclass
class Interpretability(Configuration):
    Ice: bool = True
    Icepairs: List[FeaturePair] = field(default_factory=lambda: [])
    Lime: bool = False
    Shap: ShapType = ShapType.Auto
    Shappairs: List[FeaturePair] = field(default_factory=lambda: [])
    Counterfactual: bool = False
    Anchor: bool = False


# Name collision workaround
ImputationType = Imputation
DiscretisationType = Discretisation
OutlierHandlingType = OutlierHandling
VariableTransformationType = VariableTransformation
DatetimeTransformationType = DatetimeTransformation
ScalingType = Scaling


@dataclass
class FeatureEngineeringPipeline(ImmutableConfiguration):
    Name: str = ""
    Datatype: DataType = DataType.Text
    Columns: List[str] = field(default_factory=lambda: [])
    Imputation: ImputationType = ImputationType.AutoImputer
    Encoding: CategoricalEncoding = CategoricalEncoding.NoneEncoding
    Scaling: ScalingType = ScalingType.NoScaling
    Discretisation: DiscretisationType = DiscretisationType.NoDiscretisation
    VariableTransformation: VariableTransformationType = VariableTransformationType.NoneTransform
    OutlierHandling: OutlierHandlingType = OutlierHandlingType.AutoOutlier
    DatetimeTransformation: DatetimeTransformationType = DatetimeTransformationType.NoneDatetime
    Text: TextPipelineSpec = None
    Image: ImagePipelineSpec = None
    Audio: AudioPipelineSpec = None
    Video: VideoPipelineSpec = None
    Generated: List[GeneratedColumnSpec] = field(default_factory=lambda: [])
    Custom: List[GeneratedColumnSpec] = field(default_factory=lambda: [])
    Drop: bool = False
    Passthrough: bool = False

    def to_message(self) -> MDFeatureEngineeringPipeline:
        return self.set_parent(MDFeatureEngineeringPipeline()).parent


@dataclass
class FeatureEngineeringSpec(ImmutableConfiguration):
    Pipelines: List[FeatureEngineeringPipeline] = field(default_factory=lambda: [])
    Imbalance: ImbalanceHandling = ImbalanceHandling.ImbalanceAuto
    Selection: FeatureSelection = FeatureSelection()


@dataclass
class EnsembleSpec(ImmutableConfiguration):
    Models: List[str] = field(default_factory=lambda: [])
    Estimators: List[ClassicalEstimatorSpec] = field(default_factory=lambda: [])
    Base: ClassicalEstimatorSpec = None
    Type: EnsembleType = None


TrainingSpec = Training
InterpretabilitySpec = Interpretability


@dataclass
class ModelSpec(ImmutableConfiguration):
    """
    Docstring
    """
    Owner: str = "no-one"
    VersionName: str = "v0.0.1"
    ModelVersion: str = ""
    StudyName: str = ""
    DatasetName: str = ""
    Task: TaskType = None
    Objective: Metric = None
    FeatureEngineering: FeatureEngineeringSpec = None
    Estimator: ClassicalEstimatorSpec = None
    NlpEstimator: NLPEstimatorSpec = None
    Ensemble: EnsembleSpec = None
    Training: TrainingSpec = None
    Serving: ServingSpec = None
    Tested: bool = False
    Aborted: bool = False
    Packaged: bool = False
    Published: bool = False
    Pushed: bool = False
    Reported: bool = False
    Paused: bool = False
    Profiled: bool = False
    Archived: bool = False
    Forecasted: bool = False
    Released: bool = False
    Benchmarked: bool = False
    Explained: bool = False
    Baseline: bool = False
    Flagged: bool = False
    Location: DataLocation = None
    # Forecast: ForecastSpec = None
    Compilation: CompilerSettings = None
    ActiveDeadlineSeconds: int = 600
    EstimatorType: ModelType = ModelType.Classical
    ModelClass: ModelClassType = None
    TrialID: int = 0
    Governance: GovernanceSpec = None
    Interpretability: InterpretabilitySpec = None


@dataclass
class InterpretabilityStatus(ImmutableConfiguration):
    TrainingStartTime: Time = None
    TrainingEndTime: Time = None
    ExplainerURI: str = ""
    TrainShapValuesURI: str = ""
    TestShapValuesURI: str = ""
    Importance: List[FeatureImportance] = field(default_factory=lambda: [])


@dataclass
class ModelCondition(ImmutableConfiguration):
    Type: ModelConditionType = ModelConditionType.ModelReady
    Status: ConditionStatus = ConditionStatus.ConditionUnknown
    LastTransitionTime: Time = None
    Reason: str = ""
    Message: str = ""

    def to_message(self) -> MDModelCondition:
        return self.set_parent(MDModelCondition()).parent


InterpretabilitySpec = Interpretability


@dataclass
class ModelStatus(ImmutableConfiguration):
    StartTime: Time = None
    TrainingStartTime: Time = None
    TrainingEndTime: Time = None
    TestingStartTime: Time = None
    TestingEndTime: Time = None
    EndTime: Time = None
    CvScore: float = 0
    TrainingScore: float = 0
    TestScore: float = 0
    Cost: float = 0
    Best: bool = False
    Cv: List[Measurement] = field(default_factory=lambda: [])
    Train: List[Measurement] = field(default_factory=lambda: [])
    Test: List[Measurement] = field(default_factory=lambda: [])
    Phase: ModelPhase = ModelPhase.Pending
    ReportName: str = ""
    ReportUri: str = ""
    ManifestUri: str = ""
    WeightsUri: str = ""
    LabelEncoderUri: str = ""
    LogsUri: str = ""
    ProfileUri: str = ""
    MisclassUri: str = ""
    TarUri: str = ""
    AppUri: str = ""
    ImageName: str = ""
    Importance: List[FeatureImportance] = field(default_factory=lambda: [])
    ForecastUri: str = ""
    PythonVersion: str = ""
    TrainDataset: DataLocation = None
    TestDataset: DataLocation = None
    ValidationDataset: DataLocation = None
    ObservedGeneration: int = 0
    TrainingRows: int = 0
    TestingRows: int = 0
    ValidationRows: int = 0
    FailureReason: StatusError = None
    FailureMessage: str = ""
    Progress: int = 0
    SizeInBytes: int = 0
    Latency: float = 0
    Url: str = ""
    PredictorName: str = ""
    ReleasedAt: Time = None
    TarfileHash: str = ""
    ImageHash: str = ""
    TrainingDataHash: DataHashes = None
    TrainingResources: ResourceConsumption = None
    TestingResources: ResourceConsumption = None
    TrainedBy: str = ""
    Team: str = ""
    TrainerImage: str = ""
    Endpoint: str = ""
    Logs: OutputLogs = None
    CorrelationsWithTarget: List[Correlation] = field(default_factory=lambda: [])
    TopCorrelations: List[Correlation] = field(default_factory=lambda: [])
    LastUpdated: Time = None
    # GovernanceStatus: GovernanceStatus = None
    # Interpretability: InterpretabilityStatus = None
    Conditions: List[ModelCondition] = field(default_factory=lambda: [])


@dataclass
class ModelProfile(ImmutableConfiguration):
    Name: str = ""
    Importance: List[float] = field(default_factory=lambda: [])
    Plots: List[Plot] = field(default_factory=lambda: [])


@dataclass
class AlgorithmSearchSpace(Configuration):
    Allowlist: List[ClassicEstimator] = field(default_factory=lambda: [])
    Filter: AlgorithmFilter = AlgorithmFilter.NoFilter


@dataclass
class SuccessiveHalvingOptions(Configuration):
    MaxBudget: int = 81
    EliminationRate: int = 3
    Modality: ModalityType = ModalityType.Epochs


@dataclass
class PrunerSettings(Configuration):
    Type: Pruner = Pruner.MedianPruner
    StartupTrials: int = 5
    WarmupTrials: int = 0
    MinTrials: int = 1
    IntervalTrials: int = 1
    Percentile: int = 25
    Lower: int = 0
    Upper: int = 0
    ShOptions: SuccessiveHalvingOptions = SuccessiveHalvingOptions()


SamplerType = Sampler


@dataclass
class ModelSearch(Configuration):
    Sampler: SamplerType = SamplerType.TPESearch
    Pruner: PrunerSettings = None
    MaxCost: int = 100  # Not Implemented
    MaxTime: int = 30
    MaxModels: int = 10
    MinScore: float = 0
    Trainers: int = 1
    Test: int = 1
    RetainTop: int = 10
    RetainedFor: int = 60
    SearchSpace: AlgorithmSearchSpace = AlgorithmSearchSpace()
    EarlyStopAfter: int = 0
    KeepOnlyTopModel: bool = True
    Objective: Metric = None
    Objective2: Metric = None


@dataclass
class BaselineSettings(Configuration):
    Enabled: bool = False
    Baselines: List[ClassicEstimator] = field(default_factory=lambda: [])
    All: bool = False


@dataclass
class FeatureEngineeringSearch(Configuration):
    Enabled: bool = True
    ImbalancedHandler: ImbalanceHandling = ImbalanceHandling.ImbalanceAuto
    Estimator: ClassicEstimator = ClassicEstimator.DecisionTreeClassifier  # FIXME: Should auto detect
    MaxModels: int = 2
    MaxTime: int = 3600
    MaxTrainers: int = 1
    SamplePct: int = 100
    AutoRemove: bool = True
    Reuse: bool = False
    FeatureSelectionTemplate: FeatureSelection = FeatureSelection()


@dataclass
class StudySchedule(Configuration):
    Enabled: bool = False
    StartAt: Time = None


@dataclass
class GarbageCollection(Configuration):
    CollectAtStudyEnd: bool = True
    KeepOnlyBestModelPerAlgorithm: bool = True


@dataclass
class Ensemble(Configuration):
    Enabled: bool = False
    VotingEnsemble: bool = False
    StackingEnsemble: bool = True
    Top: int = 3


@dataclass
class StudySpec(Configuration):
    VersionName: str = "v0.0.1"
    Description: str = ""
    LabRef: ObjectReference = ObjectReference("default-tenant", "default-lab")
    DatasetName: str = ""
    Task: TaskType = TaskType.AutoDetectTask
    FeSearch: FeatureEngineeringSearch = FeatureEngineeringSearch()
    Baseline: BaselineSettings = BaselineSettings()
    Search: ModelSearch = ModelSearch()
    Ensembles: Ensemble = Ensemble()
    TrainingTemplate: Training = Training()
    Schedule: StudySchedule = StudySchedule()
    Interpretability: InterpretabilitySpec = InterpretabilitySpec()
    Aborted: bool = False
    Reported: bool = True
    Paused: bool = False
    Profiled: bool = True
    ModelPublished: bool = False
    ModelImagePushed: bool = False
    ModelBenchmarked: bool = True
    ModelExplained: bool = True
    Location: DataLocation = DataLocation(BucketName="default-minio-bucket")
    Owner: str = "no-one"
    ActiveDeadlineSeconds: int = 600
    Compilation: CompilerSettings = None
    Template: bool = False
    Flagged: bool = False
    Notification: NotificationSetting = NotificationSetting()
    # ModelImage: ImageLocation = ImageLocation()
    Gc: GarbageCollection = GarbageCollection()
    Ttl: int = 0


@dataclass
class StudyCondition(Configuration):
    Type: StudyConditionType = None
    Status: ConditionStatus = None
    LastTransitionTime: Time = None
    Reason: str = ""
    Message: str = ""

    def to_message(self) -> MDStudyCondition:
        return self.set_parent(MDStudyCondition()).parent


@dataclass
class GarbageCollectionStatus(Configuration):
    Collected: int = 0


@dataclass
class StudyPhaseStatus(Configuration):
    StartTime: Time = None
    EndTime: Time = None
    Waiting: int = 0
    Running: int = 0
    Failed: int = 0
    Completed: int = 0


@dataclass
class StudyStatus(ImmutableConfiguration):
    Models: int = 0
    StartTime: Time = None
    EndTime: Time = None
    BestModel: str = ""
    BestModelScore: float = 0
    ProfileUri: str = ""
    ReportName: str = ""
    Phase: StudyPhase = StudyPhase.ModelPending
    ObservedGeneration: int = 0
    TrainDataset: DataLocation = None
    TestDataset: DataLocation = None
    ValidationDataset: DataLocation = None
    LastModelID: int = 0
    FailureReason: StatusError = None
    FailureMessage: str = ""
    TrainingRows: int = 0
    TestingRows: int = 0
    ValidationRows: int = 0
    Progress: int = 0
    BaselineModel: ClassicEstimator = None
    TrainingDataHash: DataHashes = None
    TriggeredBy: TriggerType = None
    Logs: OutputLogs = None
    FeatureEngineering: StudyPhaseStatus = None
    Baseline: StudyPhaseStatus = None
    Search: StudyPhaseStatus = None
    Ensemble: StudyPhaseStatus = None
    Test: StudyPhaseStatus = None
    Explain: StudyPhaseStatus = None
    LastUpdated: Time = None
    BestFE: FeatureEngineeringSpec = None
    Gc: GarbageCollectionStatus = None
    Conditions: List[StudyCondition] = field(default_factory=lambda: [])


@dataclass
class ReportCondition(Configuration):
    Type: ReportConditionType = None
    Status: ConditionStatus = None
    LastTransitionTime: Time = None
    Reason: str = ""
    Message: str = ""

    def to_message(self) -> MDReportCondition:
        return self.set_parent(MDReportCondition()).parent


MDReportType = ReportType


@dataclass
class ReportSpec(Configuration):
    VersionName: str = "v0.0.1"
    EntityRef: ObjectReference = None
    Location: DataLocation = DataLocation(BucketName="modela")
    ReportType: MDReportType = None
    Format: ReportFormat = ReportFormat.Pdf
    NotifierName: str = ""
    Owner: str = "no-one"
    Resources: Workload = Workload("memory-large")
    ActiveDeadlineSeconds: int = 600


@dataclass
class ReportStatus(Configuration):
    StartTime: Time = None
    EndTime: Time = None
    Phase: ReportPhase = ReportPhase.Pending
    Uri: str = ""
    ObservedGeneration: int = 0
    FailureReason: StatusError = None
    FailureMessage: str = ""
    Logs: OutputLogs = None
    LastUpdated: Time = None
    Conditions: List[ReportCondition] = field(default_factory=lambda: [])


@dataclass
class ModelAutobuilderCondition(Configuration):
    Type: ModelAutobuilderConditionType = None
    Status: ConditionStatus = None
    LastTransitionTime: Time = None
    Reason: str = ""
    Message: str = ""

    def to_message(self) -> MDModelAutobuilderCondition:
        return self.set_parent(MDModelAutobuilderCondition()).parent


DataSourceTemplate = DataSourceSpec


@dataclass
class ModelAutobuilderSpec(Configuration):
    DataProductName: str = ""
    DataProductVersionName: str = ""
    DatasourceName: str = ""
    DatasetName: str = ""
    Location: DataLocation = None
    Task: TaskType = None
    Objective: Metric = None
    TargetColumn: str = ""
    MaxTime: int = 60
    MaxModels: int = 10
    AccessMethod: AccessType = AccessType.ClusterIP
    AutoScale: bool = False
    Dataapp: bool = False
    DataSourceSpec: DataSourceSpec = None
    Trainers: int = 1
    Sampler: SamplerType = None
    Aborted: bool = False
    Owner: str = "no-one"
    Resources: Workload = None
    LabRef: ObjectReference = ObjectReference(Namespace="default-tenant", Name="default-lab")


@dataclass
class ModelAutobuilderStatus(Configuration):
    FlatFileName: str = ""
    DataSourceName: str = ""
    DatasetName: str = ""
    StudyName: str = ""
    BestModelName: str = ""
    PredictorName: str = ""
    ImageRepoName: str = ""
    Phase: ModelAutobuilderPhase = ModelAutobuilderPhase.Pending
    Rows: int = 0
    Cols: int = 0
    FileSize: int = 0
    Models: int = 0
    TrainedModels: int = 0
    BestModelScore: float = 0
    Estimator: ClassicalEstimatorSpec = None
    StartTime: Time = None
    EndTime: Time = None
    ObservedGeneration: int = 0
    FailureReason: StatusError = None
    FailureMessage: str = ""
    LastUpdated: Time = None
    Conditions: List[ModelAutobuilderCondition] = field(default_factory=lambda: [])



from enum import Enum


class ProviderName(Enum):
    AmazonAthena = "athena"
    AmazonRedshift = "redshift"
    ApacheDrill = "drill"
    ApacheDruid = "druid"
    ApacheHive = "hive"
    ApacheImpala = "impala"
    ApacheKylin = "kylin"
    ApachePinot = "pinot"
    ApacheSpark = "spark"
    AzureSqlDatabase = "azuresqldatabase"
    GcpBigQuery = "bigquery"
    ApacheCassandra = "cassandra"
    CockrouchDB = "cockrouchdb"
    Clickhouse = "clickhouse"
    Couchbase = "couchbase"
    Dremio = "dremio"
    DB2 = "db2"
    ElasticSearch = "elasticsearch"
    Exasol = "exasol"
    Kafka = "kafka"
    GoogleSheets = "googlesheets"
    GcpSpanner = "spanner"
    Greenplum = "greenplum"
    MySql = "mysql"
    MongoDb = "mongodb"
    Oracle = "oracle"
    PostgresSQL = "postgres"
    Presto = "presto"
    SAPHana = "hana"
    Snowflake = "snowflake"
    Singlestore = "singlestore"
    Sybase = "sybase"
    MSSqlServer = "mssqlserver"
    Sqlite = "sqlite"
    RabbitMQ = "rabbitmq"
    Teradata = "teradata"
    Vertica = "vertica"
    Odbc = "odbc"
    Informix = "informix"
    Neo4j = "neo4j"
    AliCloud = "alicloud"
    Azure = "azure"
    DigitalOcean = "digitalocean"
    Hetzner = "hetzner"
    Linode = "linode"
    Gcp = "gcp"
    Minio = "minio"
    Ovh = "ovh"
    Aws = "aws"
    Openstack = "openstack"
    FTP = "ftp"
    SFTP = "sftp"
    Dropbox = "dropbox"
    Box = "box"
    Dockerhub = "dockerhub"
    ImageRegistry = "image-registry"
    GitHub = "github"
    Bitbucket = "bitbucket"
    Gitlab = "gitlab"
    SmtpProvider = "smtp"
    Slack = "slack"
    Hipchat = "hipchat"
    Discord = "discord"
    PageDuty = "pagerduty"
    PushOver = "pushover"
    OpsGenie = "opsgenie"
    Webhook = "webook"
    VictorOps = "victorops"
    Flock = "flock"
    MSTeam = "msteam"
    MatterMost = "mattermost"
    Facebook = "facebook"
    Twitter = "twitter"
    UnknownProvider = "unknown"


class ConnectionCategory(Enum):
    General = "general"
    CloudProvider = "cloud"
    DockerImageRegistry = "docker-image-registry"
    Database = "database"
    Git = "git"
    Messaging = "messaging"
    ConnectionMessagerBroker = "message-broker"
    ConnectionGraphDatabase = "graph-database"
    ConnectionSocialMedia = "social-media"


class ConnectionConditionType(Enum):
    ConnectionReady = "Ready"
    ConnectionSaved = "Saved"



from abc import ABC

from github.com.metaprov.modelaapi.services.common.v1.common_pb2 import ColumnProfile as MDColumnProfile
from modela.data.common import *
from modela.infra.models import Workload, NotificationSetting, OutputLogs, GitSettings, ImageLocation
from modela.training.common import *
from modela.common import *
from modela.Configuration import *
from dataclasses import dataclass, field
from typing import List
from github.com.metaprov.modelaapi.pkg.apis.data.v1alpha1.generated_pb2 import Column as MDColumn
from github.com.metaprov.modelaapi.pkg.apis.data.v1alpha1.generated_pb2 import \
    MultiDatasetValidation as MDMultiDatasetValidation
from github.com.metaprov.modelaapi.pkg.apis.data.v1alpha1.generated_pb2 import DatasetValidation as MDDatasetValidation
from github.com.metaprov.modelaapi.pkg.apis.data.v1alpha1.generated_pb2 import \
    MultiColumnValidation as MDMultiColumnValidation
from github.com.metaprov.modelaapi.pkg.apis.data.v1alpha1.generated_pb2 import ColumnValidation as MDColumnValidation
from github.com.metaprov.modelaapi.pkg.apis.data.v1alpha1.generated_pb2 import FileValidation as MDFileValidation
from github.com.metaprov.modelaapi.pkg.apis.data.v1alpha1.generated_pb2 import DataSourceSpec as MDDataSourceSpec
from github.com.metaprov.modelaapi.pkg.apis.data.v1alpha1.generated_pb2 import RelationshipSpec as MDRelationshipSpec
from github.com.metaprov.modelaapi.pkg.apis.data.v1alpha1.generated_pb2 import Recipe as MDRecipe
from github.com.metaprov.modelaapi.pkg.apis.data.v1alpha1.generated_pb2 import RecipeSpec as MDRecipeSpec
from github.com.metaprov.modelaapi.pkg.apis.data.v1alpha1.generated_pb2 import RecipeStep as MDRecipeStep
from github.com.metaprov.modelaapi.pkg.apis.data.v1alpha1.generated_pb2 import RecipeStepParam as MDRecipeStepParam
from github.com.metaprov.modelaapi.pkg.apis.data.v1alpha1.generated_pb2 import SampleSpec as MDSampleSpec
from github.com.metaprov.modelaapi.pkg.apis.data.v1alpha1.generated_pb2 import RecipeInputSpec as MDRecipeInputSpec
from github.com.metaprov.modelaapi.pkg.apis.data.v1alpha1.generated_pb2 import RecipeOutputSpec as MDRecipeOutputSpec
from github.com.metaprov.modelaapi.pkg.apis.data.v1alpha1.generated_pb2 import Dataset as MDDataset
from github.com.metaprov.modelaapi.pkg.apis.data.v1alpha1.generated_pb2 import DatasetSpec as MDDatasetSpec
from github.com.metaprov.modelaapi.pkg.apis.data.v1alpha1.generated_pb2 import DatasetStatistics as MDDatasetStatistics
from github.com.metaprov.modelaapi.pkg.apis.data.v1alpha1.generated_pb2 import DatasetCondition as MDDatasetCondition
from github.com.metaprov.modelaapi.pkg.apis.data.v1alpha1.generated_pb2 import ColumnStatistics as MDColumnStatistics
from github.com.metaprov.modelaapi.pkg.apis.data.v1alpha1.generated_pb2 import DatasetTemplate as MDDatasetTemplate
from github.com.metaprov.modelaapi.pkg.apis.data.v1alpha1.generated_pb2 import DataLocation as MDDataLocation

from github.com.metaprov.modelaapi.pkg.apis.data.v1alpha1.generated_pb2 import \
    DataValidationResult as MDDataValidationResult
from github.com.metaprov.modelaapi.pkg.apis.data.v1alpha1.generated_pb2 import Correlation as MDCorrelation
from github.com.metaprov.modelaapi.pkg.apis.data.v1alpha1.generated_pb2 import CorrelationSpec as MDCorrelationSpec

from modela.common import Plot
from modela.util import TrackedList


@dataclass
class DataLocation(Configuration):
    Type: DataLocationType = DataLocationType.ObjectStorage
    ConnectionName: str = ""
    BucketName: str = ""
    Path: str = ""
    Table: str = ""
    Database: str = ""
    Sql: str = ""
    Topic: str = ""

    def to_message(self) -> MDDataLocation:
        return self.set_parent(MDDataLocation()).parent


@dataclass
class SampleSettings(Configuration):
    Enabled: bool = False
    Type: SamplingType = SamplingType.Random
    Rows: int = 500
    Percent: int = 100
    Filter: str = ""
    Column: str = ""

    def to_message(self) -> MDSampleSpec:
        return self.set_parent(MDSampleSpec()).parent


@dataclass
class GovernanceSpec(Configuration):
    Enabled: bool = False
    Country: str = ""
    ItReviewer: str = ""
    ComplianceReviewer: str = ""
    BusinessReviewer: str = ""


@dataclass
class CompilerSettings(Configuration):
    Enable: bool = False
    Compiler: CompilerType = CompilerType.Nothing
    Targets: List[HardwareTarget] = field(default_factory=lambda: [])


# Typename collision workaround
ColorType = Color
ImageLocationType = ImageLocation
DataLocationType = DataLocation


@dataclass
class DataProductSpec(Configuration):
    Owner: str = "no-one"
    TenantRef: ObjectReference = ObjectReference(Namespace="modela-system", Name="default-tenant")
    GitLocation: GitSettings = GitSettings()
    ImageLocation: ImageLocationType = ImageLocation()
    LabName: str = "default-lab"
    ServingSiteName: str = "default-servingsite"
    Task: TaskType = TaskType.BinaryClassification
    Description: str = ""
    DataLocation: DataLocationType = DataLocationType()
    Notification: NotificationSetting = NotificationSetting()
    Resources: Workload = Workload("general-large")
    RetriesOnFailure: int = 3
    OnCallAccountName: str = ""
    Compilation: CompilerSettings = CompilerSettings()
    ClearanceLevel: SecurityClearanceLevel = SecurityClearanceLevel.Unclassified
    Priority: PriorityLevel = PriorityLevel.Medium
    Color: ColorType = ColorType.NoColor
    Governance: GovernanceSpec = GovernanceSpec()


@dataclass
class DataProductVersionSpec(Configuration):
    ProductRef: ObjectReference = ObjectReference(Namespace="default-tenant", Name="iris-product")
    Description: str = ""
    PrevVersionName: str = ""
    Baseline: bool = False
    Owner: str = "no-one"


@dataclass
class CsvFileFormat(Configuration):
    """
    CsvFileFormat defines the file format of a raw CSV file.
    """
    ColumnDelimiter: Delimiter = Delimiter.Comma
    EscapeChar: EscapeCharacter = EscapeCharacter.DoubleQuote
    Quote: QuoteCharacter = QuoteCharacter.DoubleQuote
    CommentChars: str = "#"
    Header: bool = True
    SkipRows: int = 0
    NullValues: str = ""
    Encoding: FileEncoding = FileEncoding.Utf8
    MaxRows: int = 0
    Strict: bool = True
    Compression: CompressionType = CompressionType.Uncompressed
    HasIndexColumn: bool = False
    IndexColumn: int = 0


@dataclass
class ExcelSheetArea(Configuration):
    EntireSheet: bool = True
    FromColumn: int = 1
    ToColumn: int = 1
    FromRow: int = 1
    ToRow: int = 1


@dataclass
class ExcelNotebookFormat(Configuration):
    FirstSheetWithData: bool = False
    SheetName: str = ""
    SheetIndex: int = 0
    ColumnNameRow: int = 1
    Data: ExcelSheetArea = ExcelSheetArea()


@dataclass
class Column(Configuration):
    Name: str = ""
    Datatype: DataType = DataType.Categorical
    Format: DataDomain = DataDomain.Unknown
    DisplayName: str = ""
    Description: str = ""
    Ignore: bool = False
    Target: bool = False
    Nullable: bool = False
    Pk: bool = False
    Fk: bool = False
    MultipleOf: int = 0
    Maximum: int = 0
    ExclusiveMinimum: bool = False
    MaxLength: int = 0
    MinLength: int = 0
    Pattern: str = ""
    Required: bool = False
    Example: str = ""
    ExternalDocs: str = ""
    Enum: List[str] = field(default_factory=lambda: [])
    Ordinal: bool = False
    MaxItems: int = 0
    MinItems: int = 0
    UniqueItems: bool = False
    TimeColumn: bool = False
    Pii: bool = False
    Phi: bool = False
    PersonalData: bool = False
    Protected: bool = False
    DefaultValueNum: float = 0
    Log: bool = False
    Mu: float = 0
    Sigma: float = 0
    SkewThreshold: float = None
    DriftThreshold: float = None
    Index: bool = True
    Fold: bool = False
    Weight: bool = False
    Reserved: bool = False
    Imputation: Imputation = Imputation.AutoImputer
    Scaling: Scaling = Scaling.Auto
    Generated: bool = False
    Formula: str = ""
    Id: bool = False
    Step: float = 1

    def to_message(self) -> MDColumn:
        return self.set_parent(MDColumn()).parent


@dataclass
class TimeSeriesSchema(Configuration):
    Freq: Frequency = None
    Country: HolidayCountry = None


@dataclass
class RecommendationSchema(Configuration):
    UserIDColumn: str = "user_id"
    ItemIDColumn: str = "item_id"
    RatingColumn: str = "rating"


@dataclass
class MultiDatasetValidation(Configuration):
    Type: MultiDatasetValidationRule = None
    Datasets: List[str] = field(default_factory=lambda: [])
    Generated: bool = False

    def to_message(self) -> MDMultiDatasetValidation:
        return self.set_parent(MDMultiDatasetValidation()).parent


@dataclass
class DatasetValidation(Configuration):
    Type: DatasetValidationRule = None
    Value: float = 0
    Min: float = 0
    Max: float = 0
    ValueSet: List[str] = field(default_factory=lambda: [])
    StrictMin: bool = False
    StrictMax: bool = False
    Generated: bool = False

    def to_message(self) -> MDDatasetValidation:
        return self.set_parent(MDDatasetValidation()).parent


@dataclass
class MultiColumnValidation(Configuration):
    Type: MultiColumnValidationRule = None
    Columns: List[str] = field(default_factory=lambda: [])
    Value: float = 0
    Min: float = 0
    Max: float = 0
    ValueSet: List[str] = field(default_factory=lambda: [])
    StrictMin: bool = False
    StrictMax: bool = False
    Generated: bool = False

    def to_message(self) -> MDMultiColumnValidation:
        return self.set_parent(MDMultiColumnValidation()).parent


@dataclass
class ColumnValidation(Configuration):
    Type: ColumnValidationRule = None
    Column: str = ""
    Value: float = 0
    Min: float = 0
    Max: float = 0
    ValueSet: List[str] = field(default_factory=lambda: [])
    StrictMin: bool = False
    StrictMax: bool = False
    Generated: bool = False

    def to_message(self) -> MDColumnValidation:
        return self.set_parent(MDColumnValidation()).parent


@dataclass
class FileValidation(Configuration):
    Type: FileValidationRule = None
    BucketName: str = ""
    Column: str = ""
    Value: float = 0
    Min: float = 0
    Max: float = 0
    ValueSet: List[str] = field(default_factory=lambda: [])
    StrictMin: bool = False
    StrictMax: bool = False
    Generated: bool = False

    def to_message(self) -> MDFileValidation:
        return self.set_parent(MDFileValidation()).parent


@dataclass
class ValidationRules(Configuration):
    MultiDatasetValidations: List[MultiDatasetValidation] = field(default_factory=lambda: [])
    DatasetValidations: List[DatasetValidation] = field(default_factory=lambda: [])
    MultiColumnValidations: List[MultiColumnValidation] = field(default_factory=lambda: [])
    ColumnValidations: List[ColumnValidation] = field(default_factory=lambda: [])
    FileValidations: List[FileValidation] = field(default_factory=lambda: [])


MDTimeSeriesSchema = TimeSeriesSchema
MDRecommendationSchema = RecommendationSchema


@dataclass
class Schema(Configuration):
    Columns: List[Column] = field(default_factory=lambda: [])
    TimeSeriesSchema: MDTimeSeriesSchema = None
    RecommendationSchema: MDRecommendationSchema = None
    Validation: ValidationRules = ValidationRules()


@dataclass
class ColumnRelationship(Configuration):
    Type: str = None
    Column: str = None
    Arity: RelationshipArity = None
    RelatesTo: str = None

    def to_message(self) -> MDRelationshipSpec:
        return self.set_parent(MDRelationshipSpec()).parent


@dataclass
class CorrelationSetting(Configuration):
    Cutoff: float = 50
    Method: str = "pearson"
    Top: int = 10

    def to_message(self) -> MDCorrelationSpec:
        return self.set_parent(MDCorrelationSpec()).parent


DataSourceSchema = Schema  # workaround for type annotation bug


@dataclass
class DataSourceSpec(Configuration):
    Schema: DataSourceSchema = DataSourceSchema()
    Sample: SampleSettings = SampleSettings()
    Relationships: List[ColumnRelationship] = field(default_factory=lambda: [])
    Owner: str = "no-one"
    VersionName: str = "v0.0.1"
    Description: str = ""
    FileType: FlatFileType = FlatFileType.Csv
    Task: TaskType = TaskType.BinaryClassification
    Csvfile: CsvFileFormat = CsvFileFormat()
    ExcelNotebook: ExcelNotebookFormat = None
    DatasetType: DatasetType = DatasetType.Tabular


@dataclass
class DatasetSpec(Configuration):
    Origin: DataLocation = DataLocation()
    Location: DataLocation = DataLocation()
    Owner: str = "no-one"
    VersionName: str = "v0.0.1"
    DatasourceName: str = ""
    Description: str = ""
    DisplayName: str = ""
    Reported: bool = True
    Snapshotted: bool = False
    Validate: bool = True
    Labeled: bool = True
    Syntactic: bool = False
    SyntacticRows: int = 0
    Resources: Workload = None
    ActiveDeadlineSeconds: int = 600
    Type: DatasetType = DatasetType.Tabular
    Sample: SampleSettings = None
    Task: TaskType = None
    Notification: NotificationSetting = None
    Correlation: CorrelationSetting = None


@dataclass
class ColumnStatistics(Configuration):
    Name: str = ""
    Datatype: DataType = None
    Count: float = 0
    Distinct: int = 0
    Missing: int = 0
    PercentMissing: float = 0
    Mean: float = 0
    Stddev: float = 0
    Variance: float = 0
    Min: float = 0
    Max: float = 0
    Kurtosis: float = 0
    Skewness: float = 0
    Sum: float = 0
    Mad: float = 0
    P25: float = 0
    P50: float = 0
    P75: float = 0
    Iqr: float = 0
    Mode: str = ""
    Zeros: float = 0
    Invalid: int = 0
    Importance: float = 0
    Target: bool = False
    Ignore: bool = False
    Nullable: bool = False
    HighCardinality: bool = False
    HighCorrWithOtherFeatures: bool = False
    LowCorrWithTarget: bool = False
    HighMissingPct: bool = False
    Skewed: bool = False
    Id: bool = False
    Constant: bool = False
    Duplicate: bool = False
    Reserved: bool = False
    Outliers: int = 0
    Completeness: float = 0
    DistinctValueCount: float = 0
    MostFreqValuesRatio: float = 0
    IndexOfPeculiarity: float = 0
    CorrToTarget: float = 0

    def to_message(self) -> MDColumnStatistics:
        return self.set_parent(MDColumnStatistics()).parent


@dataclass
class ColumnProfile(Configuration):
    Name: str = ""
    Count: int = 0
    Type: str = 0
    Missing: int = 0
    PercentMissing: float = 0
    Distinct: int = 0
    Mean: float = 0
    Mode: str = ""
    Stddev: float = 0
    Variance: float = 0
    Min: float = 0
    Max: float = 0
    Kurtosis: float = 0
    Skewness: float = 0
    Sum: float = 0
    Mad: float = 0
    Zeros: float = 0
    P25: float = 0
    P50: float = 0
    P75: float = 0
    P100: float = 0
    Iqr: float = 0
    Cv: float = 0
    Top: int = 0
    Freq: int = 0
    Ignore: bool = False
    Target: bool = False
    Invalid: int = 0
    Importance: float = 0
    Nullable: bool = False
    HighCardinality: bool = False
    HighCorrWithOtherFeatures: bool = False
    LowCorrWithTarget: bool = False
    HighMissingPct: bool = False
    Skewed: bool = False
    Id: bool = False
    Constant: bool = False
    Duplicate: bool = False
    Reserved: bool = False
    Outliers: int = 0
    Completeness: float = 0
    DistinctValueCount: float = 0
    MostFreqValuesRatio: float = 0
    IndexOfPeculiarity: float = 0
    Values: List[str] = field(default_factory=lambda: [])
    CorrToTarget: float = 0

    def to_message(self) -> MDColumnProfile:
        return self.set_parent(MDColumnProfile()).parent

@dataclass
class DatasetProfile(Configuration):
    Cols: int = 0
    Rows: int = 0
    Filesize: int = 0
    Imbalance: bool = False
    Plots: List[Plot] = field(default_factory=lambda: [])
    Columns: List[ColumnProfile] = field(default_factory=lambda: [])
    Hash: str = ""


@dataclass
class Correlation(Configuration):
    Feature1: str = ""
    Feature2: str = ""
    Value: float = 0
    Method: str = ""

    def to_message(self) -> MDCorrelation:
        return self.set_parent(MDCorrelation()).parent


@dataclass
class DatasetStatistics(Configuration):
    Columns: List[ColumnStatistics] = field(default_factory=lambda: [])
    Rows: int = 0
    Cols: int = 0
    FileSize: int = 0
    CorrelationsWithTarget: List[Correlation] = field(default_factory=lambda: [])
    TopCorrelations: List[Correlation] = field(default_factory=lambda: [])

    def to_message(self) -> MDDatasetStatistics:
        return self.set_parent(MDDatasetStatistics()).parent


@dataclass
class DatasetCondition(Configuration):
    Type: DatasetCondition = DatasetCondition.Ready
    Status: ConditionStatus = ConditionStatus.ConditionUnknown
    LastTransitionTime: Time = None
    Reason: str = ""
    Message: str = ""

    def to_message(self) -> MDDatasetCondition:
        return self.set_parent(MDDatasetCondition()).parent


@dataclass
class DataValidationResult(Configuration):
    Type: str = ""
    Column: str = ""
    Error: str = ""
    Passed: bool = False

    def to_message(self) -> MDDataValidationResult:
        return self.set_parent(MDDataValidationResult()).parent


@dataclass
class DatasetStatus(ImmutableConfiguration):
    Statistics: DatasetStatistics = None
    Phase: DatasetPhase = DatasetPhase.Pending
    ReportName: str = ""
    ReportUri: str = ""
    ProfileUri: str = ""
    Imbalanced: bool = False
    ObservedGeneration: int = 0
    ValidationResults: List[DataValidationResult] = field(default_factory=lambda: [])
    LastStudyTime: Time = None
    LastNotificationTime: Time = None
    FailureReason: StatusError = None
    FailureMessage: str = ""
    Progress: int = 0
    Hash: str = ""
    Logs: OutputLogs = OutputLogs()
    DerivedFromDataset: str = ""
    LastUpdated: Time = None
    Conditions: List[DatasetCondition] = field(default_factory=lambda: [])



import time
import unittest

import pandas

from modela.ModelaException import *
from modela.data.DataSource import DataSource
from modela.data.common import DataType
from modela.data.models import DataSourceSpec, Column
from modela.server import Modela
from modela.training.common import Imputation


class Test_Modela_datasource(unittest.TestCase):
    """Tests for `modela.data.DataSource`"""

    @classmethod
    def setUpClass(cls):
        cls.modela = Modela(port_forward=True)

    @classmethod
    def tearDownClass(cls):
        cls.modela.close()

    def test_0_infer_api(self):
        with open('datasets/iris.csv') as f:
            data = f.read()

        loc = self.modela.FileService.upload_file("iris.csv", data, "default-tenant", "iris-product", "v0.0.1",
                                                  "default-minio-bucket", "datasources", "test")
        infer = self.modela.DataSources.infer("iris-product", loc)
        assert infer[0].Name == 'sepal.length'

    def test_0_create(self):
        datasource = self.modela.DataSource(namespace="iris-product", name="test")
        assert type(datasource) == DataSource
        datasource.submit(replace=True)

    def test_0_create_infer_file(self):
        datasource = self.modela.DataSource(namespace="iris-product", name="test", infer_file='datasets/iris.csv',
                                            target_column="variety")
        assert datasource.spec.Schema.Columns[0].Name == 'sepal.length'
        assert datasource.spec.Schema.Columns[4].Target

    def test_0_create_infer_bytes(self):
        with open('datasets/iris.csv', 'rb') as f:
            data = f.read()

        datasource = self.modela.DataSource(namespace="iris-product", name="test", infer_bytes=data, target_column="variety")
        assert datasource.spec.Schema.Columns[0].Name == 'sepal.length'

    def test_0_create_infer_df(self):
        df = pandas.read_csv('datasets/iris.csv')
        datasource = self.modela.DataSource(namespace="iris-product", name="test", infer_dataframe=df, target_column="variety")
        assert datasource.spec.Schema.Columns[0].Name == 'sepal.length'

    def test_1_list(self):
        assert len(self.modela.DataSources.list("iris-product")) >= 1

    def test_2_update(self):
        datasource = self.modela.DataSource(namespace="iris-product", name="test")
        datasource.spec.Schema.Columns.append(
            Column("col", DataType.Text, Imputation=Imputation.ReplaceWithMean, SkewThreshold=4, Enum=["1", "2"]))
        datasource.update()
        newds = self.modela.DataSource(namespace="iris-product", name="test")
        assert len(newds._object.spec.schema.columns) == 1

    def test_3_get(self):
        datasource = self.modela.DataSource(namespace="iris-product", name="test")
        spec = datasource.spec
        assert type(spec) == DataSourceSpec

    def test_4_delete(self):
        datasource = self.modela.DataSource(namespace="iris-product", name="test")
        datasource.delete()
        self.assertRaises(ResourceNotFoundException, self.modela.DataSources.get, "iris-product", "test")



import time
import unittest

from modela.ModelaException import *
from modela.common import PriorityLevel
from modela.data.DataProduct import DataProduct
from modela.data.common import DataType
from modela.data.models import DataProductSpec
from modela.server import Modela
from modela.training.common import Imputation


class Test_Modela_dataproduct(unittest.TestCase):
    """Tests for `modela.data.DataProduct`"""

    @classmethod
    def setUpClass(cls):
        cls.modela = Modela(port_forward=True)

    @classmethod
    def tearDownClass(cls):
        cls.modela.close()

    def test_0_create(self):
        dataproduct = self.modela.DataProduct(namespace="default-tenant", name="test")
        assert type(dataproduct) == DataProduct
        dataproduct.submit(replace=True)

    def test_1_list(self):
        assert len(self.modela.DataProducts.list("default-tenant")) > 1

    def test_2_update(self):
        time.sleep(0.3)
        dataproduct = self.modela.DataProduct(namespace="default-tenant", name="test")
        assert dataproduct.spec.Priority == PriorityLevel.Medium
        dataproduct.spec.Priority = PriorityLevel.High
        dataproduct.update()
        newdp = self.modela.DataProduct(namespace="default-tenant", name="test")
        assert newdp._object.spec.priority == "high"

    def test_3_get(self):
        dataproduct = self.modela.DataProduct(namespace="default-tenant", name="test")
        assert type(dataproduct.spec) == DataProductSpec

    def test_4_delete(self):
        dataproduct = self.modela.DataProduct(namespace="default-tenant", name="test")
        dataproduct.delete()
        self.assertRaises(ResourceNotFoundException, self.modela.DataProducts.get, "default-tenant", "test")



import grpc
from github.com.metaprov.modelaapi.pkg.apis.data.v1alpha1.generated_pb2 import DataProduct as MDDataProduct
from github.com.metaprov.modelaapi.services.dataproduct.v1.dataproduct_pb2_grpc import DataProductServiceStub
from github.com.metaprov.modelaapi.services.dataproduct.v1.dataproduct_pb2 import CreateDataProductRequest, \
    UpdateDataProductRequest, \
    DeleteDataProductRequest, GetDataProductRequest, ListDataProductsRequest

from modela.Resource import Resource
from modela.ModelaException import ModelaException
from typing import List, Union

from modela.data.models import DataProductSpec
from modela.infra.models import NotificationSetting, Workload
from modela.training.common import TaskType


class DataProduct(Resource):
    def __init__(self, item: MDDataProduct = MDDataProduct(), client=None, namespace="", name="", servingsite: str = None,
                 lab: str = None, task_type: TaskType = TaskType.BinaryClassification, default_workload: Workload = None,
                 default_bucket: str = None, notification_setting: NotificationSetting = None):
        """
        :param client: The Data Product client repository, which can be obtained through an instance of Modela.
        :param namespace: The target namespace of the resource.
        :param name: The name of the resource.
        :param servingsite: The default Serving Site of the Data Product
        :param lab: The default Lab of the Data Product
        :param task_type: The default task type of the Data Product
        :param default_workload: The default workload of the Data Product
        :param default_bucket: The default bucket used for all Data Product resources.
        :param notification_setting: The default notification settings used for all Data Product resources.
        """
        super().__init__(item, client, namespace=namespace, name=name)

    @property
    def spec(self) -> DataProductSpec:
        return DataProductSpec().copy_from(self._object.spec).set_parent(self._object.spec)

    def default(self):
        DataProductSpec().apply_config(self._object.spec)



class DataProductClient:
    def __init__(self, stub, modela):
        self.modela = modela
        self.__stub: DataProductServiceStub = stub

    def create(self, dataproduct: DataProduct) -> bool:
        request = CreateDataProductRequest()
        request.dataproduct.CopyFrom(dataproduct.raw_message)
        try:
            response = self.__stub.CreateDataProduct(request)
            return True
        except grpc.RpcError as err:
            error = err

        ModelaException.process_error(error)
        return False

    def update(self, dataproduct: DataProduct) -> bool:
        request = UpdateDataProductRequest()
        request.dataproduct.CopyFrom(dataproduct.raw_message)
        try:
            self.__stub.UpdateDataProduct(request)
            return True
        except grpc.RpcError as err:
            error = err

        ModelaException.process_error(error)
        return False

    def get(self, namespace: str, name: str) -> Union[DataProduct, bool]:
        request = GetDataProductRequest()
        request.namespace = namespace
        request.name = name
        try:
            response = self.__stub.GetDataProduct(request)
            return DataProduct(response.item, self)
        except grpc.RpcError as err:
            error = err

        ModelaException.process_error(error)
        return False

    def delete(self, namespace: str, name: str) -> bool:
        request = DeleteDataProductRequest()
        request.namespace = namespace
        request.name = name
        try:
            response = self.__stub.DeleteDataProduct(request)
            return True
        except grpc.RpcError as err:
            error = err

        ModelaException.process_error(error)
        return False

    def list(self, namespace: str) -> Union[List[DataProduct], bool]:
        request = ListDataProductsRequest()
        request.namespace = namespace
        try:
            response = self.__stub.ListDataProducts(request)
            return [DataProduct(item, self) for item in response.dataproducts.items]
        except grpc.RpcError as err:
            error = err

        ModelaException.process_error(error)
        return False





from __future__ import annotations

import json
import time

import grpc
from google.protobuf import json_format

from github.com.metaprov.modelaapi.pkg.apis.inference.v1alpha1.generated_pb2 import Predictor as MDPredictor
from github.com.metaprov.modelaapi.services.grpcinferenceservice.v1.grpcinferenceservice_pb2 import \
    PredictResultLineItem
from github.com.metaprov.modelaapi.services.predictor.v1.predictor_pb2_grpc import PredictorServiceStub
from github.com.metaprov.modelaapi.services.predictor.v1.predictor_pb2 import CreatePredictorRequest, \
    UpdatePredictorRequest, \
    DeletePredictorRequest, GetPredictorRequest, ListPredictorsRequest, PredictOneRequest

from modela.common import ObjectReference
from modela.Resource import Resource
from modela.ModelaException import ModelaException, ResourceNotFoundException
from typing import List, Union

from modela.inference.InferenceService import InferenceService
from modela.inference.common import AccessType
from modela.inference.models import ModelDeploymentSpec, PredictorSpec, PredictorStatus, PredictionResult
from modela.infra.ServingSite import ServingSite
from modela.infra.models import Workload
from modela.training.Model import Model


class Predictor(Resource):
    def __init__(self, item: MDPredictor = MDPredictor(), client=None, namespace="", name="",
                 version=Resource.DefaultVersion,
                 serving_site: Union[ObjectReference, ServingSite, str] = "default-serving-site",
                 model: Union[Model, str] = None,
                 models: List[ModelDeploymentSpec] = [],
                 port: int = 3000,
                 path: str = None,
                 access_type: AccessType = None,
                 replicas: int = 0,
                 autoscale: bool = False,
                 workload: Workload = None):
        """

        :param client: The Predictor client repository, which can be obtained through an instance of Modela.
        :param namespace: The target namespace of the resource.
        :param name: The name of the resource.
        :param version: The version of the resource.
        :param serving_site: The object reference, Serving Site object, or name under default-tenant for which the
            predictor will be deployed under.
        :param model: The Model object or name that the predictor will serve predictions for. This parameter
            will instantiate the predictor with a singe default deployment specification for the model.
        :param models: The list of model deployment specifications that will be applied to the predictor.
        :param port: The port of the predictor, which if applicable will expose the service internally or externally.
        :param path: The path that the predictor will expose a service on. If this parameter is not specified, then
            a path will be generated based on the access type and predictor name.
        :param access_type: The access type of the predictor. Documentation on how each access type exposes the prediction
            service can be found at the modela.ai Predictor resource documentation.
        :param replicas: The amount of replicas for the predictor, which if greater than zero will serve the prediction
            service on multiple pods.
        :param autoscale: If set to true, the predictor's deployment will scale based on the amount of incoming traffic
            to the service.
        :param workload: The workload specification which determines the resources which will be allocated to the
            prediction service.
        """
        super().__init__(item, client, namespace=namespace, name=name, version=version)

        if type(serving_site) == ServingSite:
            serving_site = serving_site.reference
        elif type(serving_site) == str:
            serving_site = ObjectReference(Namespace=client.modela.tenant, Name=serving_site)
        self.spec.ServingsiteRef = serving_site

        if model is not None:
            if type(model) == Model:
                model = model.name
            self.spec.Models = [ModelDeploymentSpec(ModelName=model)]
        else:
            self.spec.Models = models

        if port is not None:
            self._object.spec.port = port

        if path is not None:
            self._object.spec.path = path

        if access_type is not None:
            self._object.spec.accessType = access_type.value

        if replicas > 0:
            self._object.spec.replicas = replicas

        if workload is not None:
            self.spec.Resources = workload

    @property
    def spec(self) -> PredictorSpec:
        return PredictorSpec().copy_from(self._object.spec)

    @property
    def status(self) -> PredictorStatus:
        return PredictorStatus().copy_from(self._object.status)

    def default(self):
        PredictorSpec().apply_config(self._object.spec)

    @property
    def model(self) -> Model:
        """ Get the model associated with the Predictor """
        self.ensure_client_repository()
        return self._client.modela.Model(self.namespace, self.spec.Models[0].ModelName)

    def wait_until_ready(self):
        """ Block until the predictor service is ready to accept prediction requests. """
        payload = self.model.test_prediction
        while True:
            try:
                self.predict(payload)
            except ResourceNotFoundException:
                time.sleep(0.2)
                continue

            break

    def predict(self, predictions: str | dict | list[dict]) -> List[PredictionResult]:
        """
        Send a batch prediction request

        :param predictions: The JSON payload or dictionary(s) that contain each features name and their input value
        """
        self.ensure_client_repository()
        if type(predictions) == dict:
            predictions = json.dumps([predictions])
        elif type(predictions) == list:
            predictions = json.dumps(predictions)

        return self._client.predict(self.namespace, self.name, "", predictions)

    def connect(self, node_ip="", connect_dns=False, connect_local=False, local_port: int = None,
                tls_cert: str = None) -> InferenceService:
        """
        Connect attempts to make a connection to the gRPC inference service client associated with the Predictor.

        :rtype: InferenceService
        :param node_ip: If the Predictor's access type is Ingress or LoadBalancer, any Kubernetes node IP of the cluster
            hosting the current Modela interface must be provided. This IP can be found by executing the command
            `kubectl get nodes -o wide`
        :param connect_dns: If enabled, a connection attempt will be made using the DNS path associated
            with the Predictor (e.g. `predictor.default-serving-site.svc.cluster.local`), which can only be accessed
            from containers running inside the operational cluster.
        :param connect_local: If enabled, a connection attempt will be made with localhost.  To connect using localhost,
            you must port-forward the prediction proxy service by executing the command
            `kubectl port-forward -n default-serving-site svc/my-predictor-name 3000:3000`
        :param local_port: If connect_local is enabled, it will connect to this port.
        :return: The InferenceService client connected to the prediction proxy service
        """
        if connect_dns:
            return InferenceService(self.spec.Path, self.spec.Port)

        if connect_local:
            return InferenceService("vcap.me", local_port)

        if self.spec.AccessType == AccessType.NodePort or self.spec.AccessType == AccessType.LoadBalancer:
            return InferenceService((node_ip if node_ip != "" else self.spec.Path), self.spec.NodePort)
        elif self.spec.AccessType == AccessType.Ingress or self.spec.AccessType == AccessType.ClusterIP:
            return InferenceService(self.spec.Path, "", tls_cert=tls_cert)


class PredictorClient:
    def __init__(self, stub, modela):
        self.modela = modela
        self.__stub: PredictorServiceStub = stub

    def create(self, predictor: Predictor) -> bool:
        request = CreatePredictorRequest()
        request.predictor.CopyFrom(predictor.raw_message)
        try:
            response = self.__stub.CreatePredictor(request)
            return True
        except grpc.RpcError as err:
            error = err

        ModelaException.process_error(error)
        return False

    def update(self, predictor: Predictor) -> bool:
        request = UpdatePredictorRequest()
        request.predictor.CopyFrom(predictor.raw_message)
        try:
            self.__stub.UpdatePredictor(request)
            return True
        except grpc.RpcError as err:
            error = err

        ModelaException.process_error(error)
        return False

    def get(self, namespace: str, name: str) -> Union[Predictor, bool]:
        request = GetPredictorRequest()
        request.namespace = namespace
        request.name = name
        try:
            response = self.__stub.GetPredictor(request)
            return Predictor(response.predictor, self)
        except grpc.RpcError as err:
            error = err

        ModelaException.process_error(error)
        return False

    def delete(self, namespace: str, name: str) -> bool:
        request = DeletePredictorRequest()
        request.namespace = namespace
        request.name = name
        try:
            response = self.__stub.DeletePredictor(request)
            return True
        except grpc.RpcError as err:
            error = err

        ModelaException.process_error(error)
        return False

    def list(self, namespace: str, labels: dict = None) -> Union[List[Predictor], bool]:
        request = ListPredictorsRequest()
        request.namespace = namespace
        if labels is not None:
            request.labels.update(labels)

        try:
            response = self.__stub.ListPredictors(request)
            return [Predictor(item, self) for item in response.predictors.items]
        except grpc.RpcError as err:
            error = err

        ModelaException.process_error(error)
        return False

    def predict(self, namespace: str, name: str, fields: str, values: str) -> List[PredictionResult]:
        request = PredictOneRequest()
        request.namespace = namespace
        request.name = name
        request.fields = fields
        request.values = values

        try:
            response = self.__stub.PredictOne(request)
            output = json.loads(response.label)
            return [PredictionResult().from_dict(item) for item in output["items"]]
        except grpc.RpcError as err:
            error = err

        ModelaException.process_error(error)
        return False



import io
import json
import os.path
import random
import time

import grpc
from github.com.metaprov.modelaapi.pkg.apis.data.v1alpha1.generated_pb2 import Dataset as MDDataset
from github.com.metaprov.modelaapi.services.dataset.v1.dataset_pb2_grpc import DatasetServiceStub
from github.com.metaprov.modelaapi.services.dataset.v1.dataset_pb2 import CreateDatasetRequest, \
    UpdateDatasetRequest, \
    DeleteDatasetRequest, GetDatasetRequest, ListDatasetsRequest, GetDatasetProfileRequest
from tabulate import tabulate
from tqdm import tqdm, trange

from modela.Resource import Resource
from modela.ModelaException import ModelaException
from typing import List, Union
import pandas

from modela.data.common import *
from modela.data.DataSource import DataSource
from modela.data.models import SampleSettings, DatasetSpec, DatasetStatus, DatasetProfile
from modela.infra.models import Workload, NotificationSetting
from modela.training.Report import Report
from modela.training.common import TaskType
from modela.util import convert_size


class Dataset(Resource):
    """
    Testing dataset resource
    """
    def __init__(self, item: MDDataset = MDDataset(), client=None, namespace="", name="",
                 version=Resource.DefaultVersion,
                 gen_datasource: bool = False,
                 target_column: str = None,
                 datasource: Union[DataSource, str] = "",
                 bucket: str = "default-minio-bucket",
                 dataframe: pandas.DataFrame = None,
                 data_file: str = None,
                 data_bytes: bytes = None,
                 workload: Workload = Workload("general-large"),
                 sample=SampleSettings(),
                 task_type: TaskType = TaskType.BinaryClassification,
                 notification: NotificationSetting = None):
        """
        :param client: The Dataset client repository, which can be obtained through an instance of Modela.
        :param namespace: The target namespace of the resource.
        :param name: The name of the resource.
        :param gen_datasource: If true, a Datasource resource will be created from the uploaded dataset and applied to
            the Dataset resource.
        :param target_column: If gen_datasource is enabled, then the target column of the data source must be specified.
        :param datasource: If specified as a string, the SDK will attempt to find a Data Source resource with the given name.
            If specified as a Data Source, or if one was found with the given name, it will be applied to the Dataset.
        :param bucket: The bucket which the raw dataset data will be uploaded to.
        :param dataframe: If specified, the Pandas Dataframe will be serialized and uploaded for ingestion with the Dataset resource.
        :param data_file: If specified, the SDK will attempt read a file with the given path and will upload the
            contents of the file for ingestion with the Dataset resource.
        :param data_bytes: If specified, the SDK will upload the given raw data for ingestion with the Dataset resource.
        :param workload: The resource requirements which will be allocated for Dataset ingestion.
        :param sample: The sample settings of the dataset, which if enabled will ingest a Dataset with a portion of the uploaded data.
        :param task_type: The target task type in relation to the data being used.
        :param notification: The notification settings, which if enabled will forward events about this resource to a notifier.
        """
        self.default_resource = False
        self._profile = None
        super().__init__(item, client, namespace=namespace, name=name, version=version)
        if not self.default_resource:  # Ignore the rest of the constructor; datasets are immutable
            return

        if not gen_datasource:
            if type(datasource) != DataSource:  # Fetch the data source in case we need to read the file type
                datasource = client.modela.DataSource(namespace=namespace, name=datasource)

        if data_file is not None:
            with open(data_file, 'r') as f:
                data_bytes = f.read()
            data_file = os.path.basename(data_file)
        elif dataframe is not None:
            writer = io.BytesIO()
            if datasource.spec.FileType == FlatFileType.Csv:
                df_encoded = dataframe.to_csv(index=False)
            if datasource.spec.FileType == FlatFileType.Parquet:
                df_encoded = dataframe.to_parquet(index=False)
            elif datasource.spec.FileType == FlatFileType.Excel:
                dataframe.to_excel(writer)
                df_encoded = writer.getvalue()
            elif datasource.spec.FileType == FlatFileType.Json:
                df_encoded = dataframe.to_json()
            elif datasource.spec.FileType == FlatFileType.Feather:
                dataframe.to_feather(writer)
                df_encoded = writer.getvalue()
            else:
                raise TypeError(
                    "Pandas cannot deserialize a dataframe to the file type {0}. "
                    "Consider changing your data source file format.".format(datasource.spec.FileType))
            data_bytes = bytes(df_encoded, encoding='utf-8')

        if data_bytes is not None:
            self.spec.Origin = client.modela.FileService.upload_file(data_file or name, data_bytes,
                                                                     client.modela.tenant,
                                                                     namespace, version, bucket, "datasets", name)

        if gen_datasource:
            datasource = client.modela.DataSource(namespace=namespace, name=name + "_source", version=version,
                                                  infer_bytes=data_bytes, target_column=target_column)
            datasource.submit()

        self.spec.DatasourceName = datasource.name
        self.spec.Resources = workload
        self.spec.Task = task_type
        if sample is not None:
            self.spec.Sample = sample

        if notification is not None:
            self.spec.Notification = notification

    @property
    def spec(self) -> DatasetSpec:
        return DatasetSpec().copy_from(self._object.spec).set_parent(self._object.spec)

    @property
    def status(self) -> DatasetStatus:
        return DatasetStatus().copy_from(self._object.status)

    def default(self):
        self.default_resource = True
        DatasetSpec().apply_config(self._object.spec)

    @property
    def report(self) -> Report:
        """ Fetch the report associated with the Dataset """
        self.ensure_client_repository()
        if self._object.status.reportName != "":
            return self._client.modela.Report(namespace=self.namespace, name=self._object.status.reportName)
        else:
            print("Dataset {0} has no report.".format(self.name))

    @property
    def phase(self) -> DatasetPhase:
        """ The phase specified by the status of the Dataset """
        return self.status.Phase

    @property
    def datasource(self) -> DataSource:
        """ Fetch the DataSource object associated with the Dataset """
        self.ensure_client_repository()
        return self._client.modela.DataSource(self.namespace, self.spec.DatasourceName)

    @property
    def test_prediction(self) -> str:
        """ Generate a default prediction payload for a model that would be created with the dataset. """
        target = self.datasource.target_column.Name
        return json.dumps([{col.Name: col.Mean if col.Datatype == DataType.Number else self.datasource.column(col.Name).Enum[0]
                            for col in self.status.Statistics.Columns if col.Name != target}])

    def submit_and_visualize(self, replace: bool = False):
        """
        Submit the resource and call visualize().

        :param replace: Replace the resource if it already exists on the cluster.
        """
        self.submit(replace)
        self.visualize()

    def visualize(self):
        """
        Display a real-time visualization of the Dataset's progress
        """
        desc = tqdm(total=0, position=0, bar_format='{desc}Time Elapsed: {elapsed}')
        progress = tqdm(total=100, position=1, bar_format='{l_bar}{bar}',
                        desc=self.name, ncols=80, initial=0)

        current_status = self.status
        try:
            while True:
                self.sync()
                if current_status == self.status:
                    time.sleep(0.1)
                    continue

                current_status = self.status
                desc.set_description_str('Phase: %s | File Size: %s | Rows: %s | Columns: %s' %
                                     (self.phase.name, convert_size(self.status.Statistics.FileSize),
                                      "[Processing]" if self.status.Statistics.Rows == 0 else self.status.Statistics.Rows,
                                      "[Processing]" if len(self.status.Statistics.Columns) == 0 else len(self.status.Statistics.Columns)))
                progress.n = self.status.Progress
                progress.last_print_n = self.status.Progress
                progress.refresh()
                if self.phase in (DatasetPhase.Ready, DatasetPhase.Aborted, DatasetPhase.Failed):
                    if self.phase == DatasetPhase.Ready:
                        print("\n\n" + self.profile)
                    else:
                        progress.colour = "red"
                        progress.refresh()
                        print("\nDataset was failed or aborted: %s" % current_status.FailureMessage)

                    break

        except KeyboardInterrupt:
            pass

        desc.close()
        progress.close()

    @property
    def profile(self) -> str:
        """
        Returns an ASCII-rendered table of the Dataset's profile
        """
        profile = self.get_profile()

        table = []
        for col in profile.Columns:
            table.append([col.Name, col.Type, col.Distinct, col.Missing,
                           '{0:.3g}'.format(col.Mean), '{0:.3g}'.format(col.Stddev), col.P50, col.Min, col.Max,
                           '{0:.3g}'.format(col.CorrToTarget), '{0:.3g}'.format(col.Variance),
                           '{0:.3g}'.format(col.Skewness), '{0:.3g}'.format(col.Kurtosis)])

        table = tabulate(table, tablefmt='pretty', headers=['Column', 'Data Type', 'Distinct', 'Missing', 'Mean',
                          'Stddev', 'Median', 'Min', 'Max', 'Corr. To Target', 'Variance', 'Skewness', 'Kurtosis'])
        return table + "\n"
        # TODO: jupyter support, print all viz plots in jupyter

    @property
    def details(self) -> str:
        """
        Returns an ASCII-rendered table of the Dataset's statistics
        """
        table = []
        for col in self.status.Statistics.Columns:
            table.append([col.Name, col.Datatype.name, col.Distinct, col.Missing,
                           '{0:.3g}'.format(col.Mean), '{0:.3g}'.format(col.Stddev), col.P50, col.Min, col.Max,
                           '{0:.3g}'.format(col.Skewness), '{0:.3g}'.format(col.Kurtosis)])

        table = tabulate(table, tablefmt='pretty', headers=['Column', 'Data Type', 'Distinct', 'Missing', 'Mean',
                                                            'Stddev', 'Median', 'Min', 'Max', 'Skewness', 'Kurtosis'])
        return table + "\n"

    def get_profile(self) -> DatasetProfile:
        if self._profile:
            return self._profile
        if hasattr(self, "_client"):
            self._profile = self._client.profile(self.namespace, self.name)
            return self._profile
        else:
            raise AttributeError("Object has no client repository")


    def __repr__(self):
        out = super().__repr__()
        if len(self.status.Statistics.Columns) > 0:
            out += "\n" + self.details

        return out




class DatasetClient:
    def __init__(self, stub, modela):
        self.modela = modela
        self.__stub: DatasetServiceStub = stub

    def create(self, dataset: Dataset) -> bool:
        request = CreateDatasetRequest()
        request.dataset.CopyFrom(dataset.raw_message)
        try:
            response = self.__stub.CreateDataset(request)
            return True
        except grpc.RpcError as err:
            error = err

        ModelaException.process_error(error)
        return False

    def update(self, dataset: Dataset) -> bool:
        request = UpdateDatasetRequest()
        request.dataset.CopyFrom(dataset.raw_message)
        try:
            self.__stub.UpdateDataset(request)
            return True
        except grpc.RpcError as err:
            error = err

        ModelaException.process_error(error)
        return False

    def get(self, namespace: str, name: str) -> Union[Dataset, bool]:
        request = GetDatasetRequest()
        request.namespace = namespace
        request.name = name
        try:
            response = self.__stub.GetDataset(request)
            return Dataset(response.dataset, self)
        except grpc.RpcError as err:
            error = err

        ModelaException.process_error(error)
        return False

    def delete(self, namespace: str, name: str) -> bool:
        request = DeleteDatasetRequest()
        request.namespace = namespace
        request.name = name
        try:
            response = self.__stub.DeleteDataset(request)
            return True
        except grpc.RpcError as err:
            error = err

        ModelaException.process_error(error)
        return False

    def list(self, namespace: str) -> Union[List[Dataset], bool]:
        request = ListDatasetsRequest()
        request.namespace = namespace
        try:
            response = self.__stub.ListDatasets(request)
            return [Dataset(item, self) for item in response.datasets.items]
        except grpc.RpcError as err:
            error = err

        ModelaException.process_error(error)
        return False

    def profile(self, namespace: str, name: str) -> DatasetProfile:
        request = GetDatasetProfileRequest()
        request.namespace = namespace
        request.name = name
        try:
            response = self.__stub.GetDatasetProfile(request)
            return DatasetProfile().copy_from(response.profile)
        except grpc.RpcError as err:
            error = err

        ModelaException.process_error(error)
        return False


import time
import unittest

from modela.ModelaException import *
from modela.data.Dataset import Dataset
from modela.server import Modela
from modela import *
from modela.util import convert_size


class Test_Modela_dataset(unittest.TestCase):
    """Tests for `modela.data.DataSet`"""

    @classmethod
    def setUpClass(cls):
        cls.modela = Modela(port_forward=True)

    @classmethod
    def tearDownClass(cls):
        cls.modela.close()

    def test_0_create_datasource(self):
        datasource = self.modela.DataSource(namespace="iris-product", name="test-create-ds",
                                            infer_file='datasets/iris.csv',
                                            target_column="variety")
        datasource.submit(replace=True)
        dataset = self.modela.Dataset(namespace="iris-product", name="test-ds", data_file='datasets/iris.csv',
                                      datasource=datasource, task_type=TaskType.MultiClassification)

        dataset.submit(replace=True)


    def test_1_list(self):
        assert len(self.modela.Datasets.list("iris-product")) >= 1

    def test_1_print(self):
        print(self.modela.Datasets.list("iris-product")[0])

    def test_1_payload(self):
        print(self.modela.Datasets.list("iris-product")[0].test_prediction)

    def test_1_profile(self):
        ds = self.modela.Datasets.list("iris-product")[1]
        print(ds.profile)

    def test_2_update(self):
        dataset = self.modela.Dataset(namespace="iris-product", name="test-ds")
        dataset.set_label("test2", "e")
        dataset.update()
        newds = self.modela.Dataset(namespace="iris-product", name="test-ds")
        assert newds.has_label("test2")

    def test_3_get(self):
        dataset = self.modela.Dataset(namespace="iris-product", name="test-ds")
        spec = dataset.spec
        assert type(spec) == DatasetSpec

    def test_4_delete(self):
        dataset = self.modela.Dataset(namespace="iris-product", name="test-ds")
        # dataset.delete()
        time.sleep(0.1)
        self.assertRaises(ResourceNotFoundException, self.modela.Datasets.get, "iris-product", "test-ds")




""" These tests should be run by an end-user. """
def test_viz():
    modela = Modela("localhost", 3000)
    datasource = modela.DataSource(namespace="iris-product", name="iris")
    dataset = modela.Dataset(namespace="iris-product", name="test-ds-5", data_file='tests/datasets/iris.csv',
                             datasource=datasource, task_type=TaskType.MultiClassification)


    dataset.submit_and_visualize(replace=True)
    modela.close()




from dataclasses import dataclass, field
from modela.Configuration import *
from github.com.metaprov.modelaapi.pkg.apis.catalog.v1alpha1.generated_pb2 import ResourceSpec as MDResourceSpec
from github.com.metaprov.modelaapi.pkg.apis.catalog.v1alpha1.generated_pb2 import NotificationSpec as MDNotificationSpec
from github.com.metaprov.modelaapi.pkg.apis.catalog.v1alpha1.generated_pb2 import Logs as MDLogs
from github.com.metaprov.modelaapi.pkg.apis.catalog.v1alpha1.generated_pb2 import ContainerLog as MDContainerLog


@dataclass
class NotificationSetting(Configuration):
    ErrorTTL: int = 3600
    SuccessTTL: int = 3600
    NotifierName: str = ""
    # FIXME: Implement dict types: Selector: dict = field(default_factory=lambda: {})

@dataclass
class Workload(Configuration):
    WorkloadClassName: str = ""
    Enable: bool = False


@dataclass
class ContainerLog(Configuration):
    Job: str = ""
    Container: str = ""
    Key: str = ""

    def to_message(self) -> MDContainerLog:
        return self.set_parent(MDContainerLog()).parent

@dataclass
class OutputLogs(Configuration):
    BucketName: str = ""
    Containers: List[ContainerLog] = field(default_factory=lambda: [])

    def to_message(self) -> MDLogs:
        return self.set_parent(MDLogs()).parent


@dataclass
class GitSettings(Configuration):
    GitConnectionName: str = ""
    Url: str = ""
    Branch: str = ""
    Private: bool = False


@dataclass
class ImageLocation(Configuration):
    Name: str = ""
    RegistryConnectionName: str = ""
